@@ -1,16 +1,37 @@
+import collections
+import hashlib
+import logging
 import numpy as np
 import os
 import platform
+import requests
+import sys
+import tarfile
+import tempfile
 import xdg
 
 import py4dgeo._py4dgeo as _py4dgeo
 
 
+# The current data archive URL
+TEST_DATA_ARCHIVE = "https://github.com/3dgeo-heidelberg/py4dgeo-test-data/releases/download/2022-06-22/data.tar.gz"
+TEST_DATA_CHECKSUM = "7b459de25c210da3c4815b19b09647dfc73cff0c692954cd7ec29b166ec2c42a"
+
+# The directory where to place
+_test_data_dir = tempfile.TemporaryDirectory()
+
+
 class Py4DGeoError(Exception):
-    pass
+    def __init__(self, msg, loggername="py4dgeo"):
+        # Initialize the base class
+        super().__init__(msg)
+
+        # Also write the message to the error stream
+        logger = logging.getLogger(loggername)
+        logger.error(self)
 
 
-def find_file(filename):
+def find_file(filename, fatal=True):
     """Find a file of given name on the file system.
 
     This function is intended to use in tests and demo applications
@@ -22,10 +43,14 @@ def find_file(filename):
     * If an absolute filename is given, it is used
     * Check whether the given relative path exists with respect to the current working directory
     * Check whether the given relative path exists with respect to the specified XDG data directory (e.g. through the environment variable :code:`XDG_DATA_DIRS`).
+    * Check whether the given relative path exists in downloaded test data.
 
     :param filename:
         The (relative) filename to search for
     :type filename: str
+    :param fatal:
+        Whether not finding the file should be a fatal error
+    :type fatal: bool
     :return: An absolute filename
     """
 
@@ -44,14 +69,21 @@ def find_file(filename):
         for xdg_dir in xdg.xdg_data_dirs():
             candidates.append(os.path.join(xdg_dir, filename))
 
+    # Use the temporary directory with test data (only contains something)
+    # iff ensure_test_data_availability was called.
+    candidates.append(os.path.join(_test_data_dir.name, filename))
+
     # Iterate through the list to check for file existence
     for candidate in candidates:
         if os.path.exists(candidate):
             return candidate
 
-    raise FileNotFoundError(
-        f"Cannot locate file {filename}. Tried the following locations: {', '.join(candidates)}"
-    )
+    if fatal:
+        raise FileNotFoundError(
+            f"Cannot locate file {filename}. Tried the following locations: {', '.join(candidates)}"
+        )
+
+    return filename
 
 
 class MemoryPolicy(_py4dgeo.MemoryPolicy):
@@ -135,7 +167,19 @@ def make_contiguous(arr: np.ndarray):
     return np.copy(arr, order="C")
 
 
-def as_double_precision(arr: np.ndarray):
+def _as_dtype(arr, dtype, policy_check):
+    if np.issubdtype(arr.dtype, dtype):
+        return arr
+
+    if policy_check and not memory_policy_is_minimum(MemoryPolicy.MINIMAL):
+        raise Py4DGeoError(
+            f"py4dgeo expected an input of type {np.dtype(dtype).name}, but got {np.dtype(arr.dtype).name}. Current memory policy forbids automatic cast."
+        )
+
+    return np.asarray(arr, dtype=dtype)
+
+
+def as_double_precision(arr: np.ndarray, policy_check=True):
     """Ensure that a numpy array is double precision
 
     This is a no-op if the array is already double precision and makes a copy
@@ -144,16 +188,7 @@ def as_double_precision(arr: np.ndarray):
     :param arr: The numpy array
     :type arr: np.ndarray
     """
-
-    if arr.dtype.name == "float64":
-        return arr
-
-    if not memory_policy_is_minimum(MemoryPolicy.MINIMAL):
-        raise Py4DGeoError(
-            "py4dgeo only operates on double precision input. Please provide it or set the memory policy to MINIMAL or above."
-        )
-
-    return np.asarray(arr, dtype=np.float64)
+    return _as_dtype(arr, np.float64, policy_check)
 
 
 def set_num_threads(num_threads: int):
@@ -183,3 +218,50 @@ def get_num_threads():
     except AttributeError:
         # The C++ library was built without OpenMP!
         return 1
+
+
+def append_file_extension(filename, extension):
+    """Append a file extension if and only if the original filename has none"""
+
+    _, ext = os.path.splitext(filename)
+    if ext == "":
+        return f"{filename}.{extension}"
+    else:
+        return filename
+
+
+def is_iterable(obj):
+    """Whether the object is an iterable (excluding a string)"""
+    return isinstance(obj, collections.abc.Iterable) and not isinstance(obj, str)
+
+
+def copy_test_data(target):
+    """Download test data and copy it into the working directory"""
+
+    # Create temporary directory for the download
+    with tempfile.TemporaryDirectory() as tmp:
+        archive = requests.get(TEST_DATA_ARCHIVE).content
+        checksum = hashlib.sha256(archive).hexdigest()
+        if checksum != TEST_DATA_CHECKSUM:
+            raise ValueError("Checksum for test data archive failed.")
+
+        archive_file = os.path.join(tmp, "data.tar.gz")
+        with open(archive_file, "wb") as tar:
+            tar.write(archive)
+
+        with tarfile.open(archive_file, "r:gz") as tar:
+            tar.extractall(path=target)
+
+
+def copy_test_data_entrypoint():
+    # Define the target directory
+    target = os.getcwd()
+    if len(sys.argv) > 1:
+        target = sys.argv[1]
+
+    copy_test_data(target)
+
+
+def ensure_test_data_availability():
+    """Ensure the availability of test data for notebooks etc."""
+    copy_test_data(_test_data_dir.name)
